import sqlalchemy.orm
import logging

log = logging.getLogger(__name__)

from lokiunimore.config import SQLALCHEMY_DATABASE_URL


log.debug(f"Creating SQLAlchemy engine for: {__name__}")
engine = sqlalchemy.create_engine(SQLALCHEMY_DATABASE_URL.__wrapped__)
log.debug(f"Created SQLAlchemy engine for: {__name__}")


__all__ = (
    "engine",
)



import nio
import logging
import sqlalchemy.orm
import functools

log = logging.getLogger(__name__)

from lokiunimore.config import MATRIX_SKIP_EVENTS
from lokiunimore.matrix.database import engine
from lokiunimore.sql.tables import MatrixProcessedEvent


def filter_processed_events(f):
    """
    Decorator applicable to a :mod:`nio` callback to filter incoming events whose IDs have been marked in the database as *already processed*, and marking events successfully processed by the function as "processed".
    """

    @functools.wraps(f)
    async def wrapped(room: nio.MatrixRoom, event: nio.Event):
        log.debug(f"Checking if event should be processed: {event.event_id}")
        if should_process_event(event):
            if not MATRIX_SKIP_EVENTS.__wrapped__:
                log.debug(f"Processing event: {event.event_id}")
                await f(room, event)
            else:
                log.debug(f"Skipping event due to MATRIX_SKIP_EVENTS: {event.event_id}")
            log.debug(f"Marking event as processed: {event.event_id}")
            mark_event_as_processed(event)
        else:
            log.debug(f"Skipping already processed event: {event.event_id}")

    return wrapped


def should_process_event(event: nio.Event) -> bool:
    """
    Check if the event should be processed.

    :param event: The event to check.
    :return: :data:`True` if the event should be processed, :data:`False` otherwise.
    """

    with sqlalchemy.orm.Session(bind=engine) as session:
        session: sqlalchemy.orm.Session
        return session.query(MatrixProcessedEvent).get(event.event_id) is None


def mark_event_as_processed(event: nio.Event):
    """
    Mark the event as processed, preventing it from being processed again.

    :param event: The event to mark.
    :return: The created and committed :class:`lokiunimore.sql.tables.MatrixProcessedEvent` instance.
    """

    with sqlalchemy.orm.Session(bind=engine) as session:
        session: sqlalchemy.orm.Session
        mpe = MatrixProcessedEvent(id=event.event_id)
        session.add(mpe)
        session.commit()
        return mpe


__all__ = (
    "filter_processed_events",
)



import logging
import sqlalchemy.orm
import nio.store.database

log = logging.getLogger(__name__)

from lokiunimore.matrix.database import engine
from lokiunimore.matrix.templates.messages import WELCOME_MESSAGE_TEXT, WELCOME_MESSAGE_HTML, SUCCESS_MESSAGE_TEXT, SUCCESS_MESSAGE_HTML, GOODBYE_MESSAGE_TEXT, GOODBYE_MESSAGE_HTML, UNLINK_MESSAGE_TEXT, UNLINK_MESSAGE_HTML
from lokiunimore.matrix.client_class import ExtendedAsyncClient, RequestError
from lokiunimore.matrix.event_filtering import filter_processed_events
from lokiunimore.config import MATRIX_HOMESERVER, MATRIX_USER_ID, MATRIX_STORE_DIR, MATRIX_PUBLIC_SPACE_ID, MATRIX_PRIVATE_SPACE_ID, FLASK_BASE_URL
from lokiunimore.sql.tables import MatrixUser


client = ExtendedAsyncClient(
    homeserver=MATRIX_HOMESERVER.__wrapped__,
    user=MATRIX_USER_ID.__wrapped__,
    store_path=str(MATRIX_STORE_DIR.__wrapped__),
    config=nio.AsyncClientConfig(
        store_sync_tokens=True,
        store=nio.store.database.DefaultStore,
    ),
)
"""
The bot's Matrix client, powered by :mod:`nio`.
"""


@filter_processed_events
async def on_member(room: nio.MatrixRoom, event: nio.Event):
    """
    Callback for when a :class:`nio.RoomMemberEvent` is received.
    """

    # Filters allow us to determine the event type in a better way
    event: nio.InviteMemberEvent | nio.RoomMemberEvent

    # Catch events about myself immediately
    if event.state_key == client.user_id:
        # If I'm invited to a room, join it
        if event.membership == "invite":
            await join_when_invited(room)

    elif event.membership == "join":
        # If somebody joins the parent space, notify them of my presence and send them their profile URL
        if room.room_id == MATRIX_PUBLIC_SPACE_ID:
            await handle_public_space_joiner(event.state_key)
        # If somebody joins the child space, notify them of the successful login
        elif room.room_id == MATRIX_PRIVATE_SPACE_ID:
            await handle_private_space_joiner(event.state_key)

    elif event.membership == "leave":
        # If somebody leaves the parent space, remove them from all subrooms, delete their account, and finally notify them
        if room.room_id == MATRIX_PUBLIC_SPACE_ID:
            await handle_public_space_leaver(event.state_key)
        # If somebody leaves the child space, remove them from all subrooms, delete their linking, and finally notify them
        elif room.room_id == MATRIX_PRIVATE_SPACE_ID:
            await handle_private_space_leaver(event.state_key)

    elif event.membership == "ban":
        # TODO: If somebody is banned from the parent space... propagate the ban?
        if room.room_id == MATRIX_PUBLIC_SPACE_ID:
            await handle_public_space_leaver(event.state_key)
        # TODO: If somebody is banned from the child space... propagate the ban?
        elif room.room_id == MATRIX_PRIVATE_SPACE_ID:
            await handle_private_space_leaver(event.state_key)


async def join_when_invited(room: nio.MatrixRoom):
    log.debug(f"Received invite to: {room.room_id}")
    await client.join(room.room_id)
    log.info(f"Accepted invite to: {room.room_id}")


async def handle_public_space_joiner(user_id: str):
    log.debug(f"User joined public space: {user_id}")

    log.debug(f"Creating MatrixUser for: {user_id}")
    with sqlalchemy.orm.Session(bind=engine) as session:
        session: sqlalchemy.orm.Session
        matrix_user = MatrixUser(id=user_id)
        matrix_user = session.merge(matrix_user)
        session.commit()
        log.debug(f"Created MatrixUser for: {user_id}")
        token = matrix_user.token

    log.debug(f"Notifying user of the account creation: {user_id}")
    formatting = dict(
        username_text=client.user_id,
        username_html=await client.mention_html(user_id),
        base_url=FLASK_BASE_URL,
        token=token,
    )
    await client.room_send_message_html(
        await client.find_or_create_pm_room(user_id),
        text=WELCOME_MESSAGE_TEXT.format(**formatting),
        html=WELCOME_MESSAGE_HTML.format(**formatting)
    )
    log.debug(f"Notified user of the account creation: {user_id}")

    log.info(f"Handled joiner of public space: {user_id}")


async def handle_private_space_joiner(user_id: str):
    log.info(f"User joined private space: {user_id}")

    log.debug(f"Setting MatrixUser as joined for: {user_id}")
    with sqlalchemy.orm.Session(bind=engine) as session:
        session: sqlalchemy.orm.Session
        matrix_user: MatrixUser = session.query(MatrixUser).get(user_id)
        matrix_user.joined_private_space = True
        session.commit()
        log.debug(f"Set MatrixUser as joined for: {user_id}")
        token = matrix_user.token

    log.debug(f"Notifying user of the account link: {user_id}")
    formatting = dict(
        base_url=FLASK_BASE_URL,
        token=token,
    )
    await client.room_send_message_html(
        await client.find_or_create_pm_room(user_id),
        text=SUCCESS_MESSAGE_TEXT.format(**formatting),
        html=SUCCESS_MESSAGE_HTML.format(**formatting)
    )
    log.debug(f"Notified user of the account link: {user_id}")

    log.info(f"Handled joiner of private space: {user_id}")


async def handle_public_space_leaver(user_id: str):
    log.info(f"User left public space: {user_id}")

    log.debug(f"Deleting MatrixUser for: {user_id}")
    with sqlalchemy.orm.Session(bind=engine) as session:
        session: sqlalchemy.orm.Session
        matrix_user: MatrixUser = session.query(MatrixUser).get(user_id)
        if matrix_user is None:
            log.warning(f"User left public space without having a pre-existent record in the db: {user_id}")
        else:
            if matrix_user.account is not None and len(matrix_user.account.matrix_users) == 1:
                session.delete(matrix_user.account)
            session.delete(matrix_user)
            session.commit()
            log.debug(f"Deleted MatrixUser for: {user_id}")

    log.debug(f"Notifying user of the account deletion: {user_id}")
    await client.room_send_message_html(
        await client.find_or_create_pm_room(user_id),
        text=GOODBYE_MESSAGE_TEXT,
        html=GOODBYE_MESSAGE_HTML
    )
    log.debug(f"Notified user of the account deletion: {user_id}")

    log.debug(f"Finding room hierarchy of the public space...")
    hierarchy = await client.room_hierarchy(MATRIX_PUBLIC_SPACE_ID.__wrapped__, max_depth=9, suggested_only=False)

    log.debug(f"Removing public space leaver from {len(hierarchy)} rooms: {user_id}")
    success_count = 0
    for room in hierarchy:
        room_id = room["room_id"]
        try:
            await client.room_kick(room_id=room_id, user_id=user_id, reason="Left parent space")
        except RequestError as e:
            log.warning(f"Could not remove public space leaver {user_id} from {room_id}: {e!r}")
        else:
            success_count += 1
    log.debug(f"Removed public space leaver from {success_count} rooms: {user_id}")

    log.info(f"Handled leaver of public space: {user_id}")


async def handle_private_space_leaver(user_id: str):
    log.debug(f"User left private space: {user_id}")

    log.debug(f"Unlinking account for: {user_id}")
    with sqlalchemy.orm.Session(bind=engine) as session:
        session: sqlalchemy.orm.Session
        matrix_user: MatrixUser = session.query(MatrixUser).get(user_id)
        if matrix_user is None:
            log.warning(f"User left private space without having a pre-existent record in the db: {user_id}")
        elif matrix_user.account is None:
            log.warning(f"User left private space without having a linked account in the db: {user_id}")
            token = matrix_user.token
        else:
            if len(matrix_user.account.matrix_users) == 1:
                session.delete(matrix_user.account)
            matrix_user.account = None
            matrix_user.joined_private_space = False
            session.commit()
            log.debug(f"Unlinked account for: {user_id}")
            token = matrix_user.token

    log.debug(f"Notifying user of the account unlinking: {user_id}")
    formatting = dict(
        base_url=FLASK_BASE_URL,
        token=token,
    )
    await client.room_send_message_html(
        await client.find_or_create_pm_room(user_id),
        text=UNLINK_MESSAGE_TEXT.format(**formatting),
        html=UNLINK_MESSAGE_HTML.format(**formatting)
    )
    log.debug(f"Notified user of the account unlinking: {user_id}")

    log.debug(f"Finding room hierarchy of the private space...")
    hierarchy = await client.room_hierarchy(MATRIX_PRIVATE_SPACE_ID.__wrapped__, max_depth=9, suggested_only=False)

    log.debug(f"Removing private space leaver from {len(hierarchy)} rooms: {user_id}")
    success_count = 0
    for room in hierarchy:
        room_id = room["room_id"]
        try:
            await client.room_kick(room_id=room_id, user_id=user_id, reason="Left child space")
        except RequestError as e:
            log.warning(f"Could not remove private space leaver {user_id} from {room_id}: {e}")
        else:
            success_count += 1
    log.debug(f"Removed private space leaver from {success_count} rooms: {user_id}")

    log.info(f"Handled leaver of private space: {user_id}")


log.debug(f"Registering client callbacks on: {client}")

# noinspection PyTypeChecker
client.add_event_callback(on_member, nio.InviteMemberEvent)
client.add_event_callback(on_member, nio.RoomMemberEvent)

log.debug(f"Registered client callbacks on: {client}")



import flask
import hmac
import hashlib
import requests

from lokiunimore.utils.device_names import generate_device_name


class MatrixClientExtension:
    """
    Flask extension providing a extremely simple, :mod:`requests`-based, synchronous Matrix client.
    """

    def __init__(self, app: flask.Flask = None):
        if app is not None:
            self.init_app(app)

        token = hmac.new(
            key=app.config["MATRIX_USER_SECRET"].encode("utf8"),
            msg=app.config["MATRIX_USER_ID"].encode("utf8"),
            digestmod=hashlib.sha512
        ).hexdigest()

        response = requests.post(f"{app.config['MATRIX_HOMESERVER']}/_matrix/client/v3/login", json={
            "type": "com.devture.shared_secret_auth",
            "identifier": {
                "type": "m.id.user",
                "user": app.config["MATRIX_USER_ID"],
            },
            "token": token,
            "initial_device_display_name": generate_device_name(__name__),
        })
        response.raise_for_status()
        response = response.json()

        self.access_token = response["access_token"]
        self.device_id = response["device_id"]

    def init_app(self, app: flask.Flask):
        pass

    def room_invite(self, room_id: str, user_id: str):
        """
        Send an invite to a Matrix user for a certain room.

        :param room_id: The room to invite the user to.
        :param user_id: The user to invite.
        """

        response = requests.post(
            f"{flask.current_app.config['MATRIX_HOMESERVER']}/_matrix/client/v3/rooms/{flask.current_app.config['MATRIX_PRIVATE_SPACE_ID']}/invite",
            json={
                "reason": "Account linked",
                "user_id": user_id,
            },
            headers={
                "Authorization": f"Bearer {self.access_token}",
            }
        )
        response.raise_for_status()



"""
This module defines some extensions for :mod:`nio`, as it is currently missing some methods that :mod:`lokiunimore` desperately needs.
"""

import aiohttp
import nio
import nio.crypto
import logging
import hashlib
import hmac
import typing as t

T = t.TypeVar("T")
Something = t.TypeVar("Something")
log = logging.getLogger(__name__)

from lokiunimore.utils.device_names import generate_device_name


class RequestError(Exception):
    """
    A request to the Matrix homeserver failed.
    """

    def __init__(self, response: nio.responses.ErrorResponse):
        super().__init__()
        self.response: nio.responses.ErrorResponse = response


# noinspection PyProtectedMember
class ExtendedAsyncClient(nio.AsyncClient):
    """
    An :class:`~nio.AsyncClient` with some extra features to be upstreamed some day.
    """

    def __repr__(self):
        return f"<{self.__class__.__qualname__} for {self.user} at {self.homeserver}>"

    def _send(
            self,
            response_class: t.Type[T],
            method: str,
            path: str,
            data: t.Union[None, str, nio.crypto.AsyncDataT] = None,
            response_data: t.Optional[t.Tuple[t.Any, ...]] = None,
            content_type: t.Optional[str] = None,
            trace_context: t.Optional[t.Any] = None,
            data_provider: t.Optional[nio.DataProvider] = None,
            timeout: t.Optional[float] = None,
            content_length: t.Optional[int] = None,
    ) -> T:
        """
        Override :meth:`nio.client.AsyncClient._send` with something that actually raises errors instead of returning them.

        .. danger::

            SUPER UNTESTED. PLEASE WORK. PLEASE DO. PLEASE DO NOT EXPLODE IN PRODUCTION.
        """
        result = super()._send(
            response_class=response_class,
            method=method,
            path=path,
            data=data,
            response_data=response_data,
            content_type=content_type,
            trace_context=trace_context,
            data_provider=data_provider,
            timeout=timeout,
            content_length=content_length,
        )

        if isinstance(result, nio.responses.ErrorResponse):
            log.warning(f"{method} {path} errored: {result!r}")
            raise RequestError(result)

        return result

    async def register_with_shared_secret(self, shared_secret: str, username: str, displayname: str, password: str):
        """
        Given the homeserver's shared secret, perform a `shared-secret registration <https://matrix-org.github.io/synapse/latest/admin_api/register_api.html#shared-secret-registration>`_.

        :param shared_secret: The shared secret to register with.
        :param username: The username of the user to register.
        :param displayname: The displayname of the user to register.
        :param password: The password of the user to register.
        :raises aiohttp.ClientResponseError: If a request is not successful.

        .. note:: ``shared_secret`` is not the same shared secret of :meth:`.login_with_shared_secret`.
        """

        log.debug(f"Registering via shared-secret registration: {username}")

        path = nio.Api._build_path(
            ["v1", "register"],
            base_path="/_synapse/admin"
        )

        nonce: aiohttp.ClientResponse = await self.send("GET", path)
        nonce.raise_for_status()
        nonce: dict[str, ...] = await nonce.json()
        nonce: str = nonce["nonce"]

        mac = hmac.new(shared_secret.encode("utf8"), digestmod=hashlib.sha1)
        mac.update(nonce.encode("utf8"))
        mac.update(b"\0")
        mac.update(username.encode("utf8"))
        mac.update(b"\0")
        mac.update(password.encode("utf8"))
        mac.update(b"\0")
        mac.update(b"notadmin")

        registration: aiohttp.ClientResponse = await self.send("POST", path, nio.Api.to_json({
            "nonce": nonce,
            "username": username,
            "displayname": displayname,
            "password": password,
            "admin": False,
            "mac": mac.hexdigest()
        }))
        registration.raise_for_status()

        log.info(f"Registered via shared-secret registration: {username}")

    async def login_with_shared_secret(self, shared_secret: str) -> None:
        """
        Given a shared secret, perform a login via `com.devture.shared_secret_auth <https://github.com/devture/matrix-synapse-shared-secret-auth>`_.

        :param shared_secret: The shared secret to login with.

        .. note:: ``shared_secret`` is not the same shared secret of :meth:`.register_with_shared_secret`.
        """

        log.debug(f"Logging in as {self.user} with a shared secret...")

        token = hmac.new(key=shared_secret.encode("utf8"), msg=self.user.encode("utf8"), digestmod=hashlib.sha512).hexdigest()

        await self.login_raw({
            "type": "com.devture.shared_secret_auth",
            "identifier": {
                "type": "m.id.user",
                "user": self.user,
            },
            "token": token,
            "initial_device_display_name": generate_device_name(__name__),
        })

        log.debug(f"Login successful!")

    async def room_hierarchy(self, room_id: str, max_depth: int, suggested_only: bool) -> list[dict]:
        """
        Given a space, get the room hierarchy.

        :param room_id: The id of the space to get the hierarchy of.
        :param max_depth: The max depth of subspaces to recurse into.
        :param suggested_only: Whether only suggested rooms should be returned.
        :return: 
        """

        log.debug(f"Getting room hierarchy for: {room_id!r}")

        current = None
        rooms = []

        while True:
            path = ["rooms", room_id, "hierarchy"]
            query = {
                "max_depth": str(max_depth),
                "suggested_only": str(suggested_only).lower(),
            }
            if current:
                query["from"] = current

            result: aiohttp.ClientResponse = await self.send(
                "GET",
                nio.Api._build_path(path, query, base_path="/_matrix/client/v1"),
                headers={
                    "Authorization": f"Bearer {self.access_token}"
                }
            )
            result: dict = await result.json()

            rooms = [*rooms, *result["rooms"]]

            current = result.get("next_batch")
            if not current:
                break

        log.debug(f"Successfully retrieved a hierarchy of {len(rooms)} rooms!")

        return rooms

    async def find_or_create_pm_room(self, user_id: str) -> str:
        """
        Find the first available private message room with the given user, or create one if none exist.

        :param user_id: The user to message.
        :returns: The room id of the created room.
        """

        log.debug(f"Sliding into {user_id}'s PMs...")

        for room in self.rooms.values():
            is_dm = "m.direct" in room.tags
            is_group = room.is_group
            has_two_users = len(room.users) + len(room.invited_users) == 2
            contains_user_id = user_id in room.users.keys() or user_id in room.invited_users.keys()

            if (is_dm or is_group) and has_two_users and contains_user_id:
                log.debug(f"Found PM room for: {user_id}")
                return room.room_id
        else:
            log.debug(f"Creating new PM room for: {user_id}")
            response = await self.room_create(invite=[user_id], is_direct=True)
            if isinstance(response, nio.RoomCreateResponse):
                log.info(f"Created new PM room for: {user_id}")
                return response.room_id
            else:
                raise Exception("Failed to slide into an user's PMs.")

    async def room_send_message_html(self, room_id: str, text: str, html: str) -> Something:
        """
        Send an HTML message with a text fallback.

        :param room_id: The ID of the room to send the message in.
        :param text: The plain text fallback of the message.
        :param html: The HTML message.
        :return: I have no idea, whatever :meth:`~nio.client.AsyncClient.room_send` does.
        """

        return await self.room_send(room_id, "m.room.message", {
            "msgtype": "m.notice",
            "format": "org.matrix.custom.html",
            "formatted_body": html,
            "body": text,
        })

    async def mention_html(self, user_id: str) -> str:
        """
        Create a rich HTML mention.

        To provide a text fallback for the HTML mention, use the ``user_id``.

        :param user_id: The user to mention.
        :return: The HTML string.
        """

        display_name = await self.get_displayname(user_id)

        # language=html
        return f"""<a href="https://matrix.to/#/{user_id}">{display_name}</a>"""



import coloredlogs
import logging
import asyncio

coloredlogs.install(
    logger=logging.getLogger("lokiunimore"),
    level="DEBUG",
    fmt="{asctime} | {name:<32} | {levelname:>8} | {message}",
    style="{",
    level_styles=dict(
        debug=dict(color="white"),
        info=dict(color="cyan"),
        warning=dict(color="yellow"),
        error=dict(color="red"),
        critical=dict(color="red", bold=True),
    ),
    field_styles=dict(
        asctime=dict(color='magenta'),
        levelname=dict(color='blue', bold=True),
        name=dict(color='blue'),
    ),
    isatty=True,
)

from lokiunimore.matrix.client import client
from lokiunimore.config import MATRIX_USER_SECRET

loop = asyncio.new_event_loop()
log = logging.getLogger(__name__)


async def main():
    await client.login_with_shared_secret(MATRIX_USER_SECRET)
    await client.sync_forever(60_000, full_state=True)


async def cleanup():
    await client.logout()


try:
    loop.run_until_complete(main())
finally:
    loop.run_until_complete(cleanup())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lokiunimore',
 'lokiunimore.config',
 'lokiunimore.matrix',
 'lokiunimore.matrix.templates',
 'lokiunimore.sql',
 'lokiunimore.utils',
 'lokiunimore.web',
 'lokiunimore.web.extensions']

package_data = \
{'': ['*'], 'lokiunimore.web': ['static/*', 'templates/*']}

install_requires = \
['Authlib>=1.0.1,<2.0.0',
 'Flask-SQLAlchemy>=2.5.1,<3.0.0',
 'Flask>=2.2.2,<3.0.0',
 'cfig[cli]>=0.3.0,<0.4.0',
 'coloredlogs>=15.0.1,<16.0.0',
 'gunicorn>=20.1.0,<21.0.0',
 'matrix-nio[e2e]>=0.19.0,<0.20.0',
 'psycopg2>=2.9.3,<3.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'requests>=2.28.1,<3.0.0']

entry_points = \
{'console_scripts': ['lokiunimore = lokiunimore.__main__:main']}

setup_kwargs = {
    'name': 'lokiunimore',
    'version': '0.1.0',
    'description': 'Matrix room gatekeeper bot for the unofficial Unimore space',
    'long_description': '# `lokiunimore`\n\nMatrix room gatekeeper bot for the unofficial Unimore space\n',
    'author': 'Stefano Pigozzi',
    'author_email': 'me@steffo.eu',
    'maintainer': 'Stefano Pigozzi',
    'maintainer_email': 'me@steffo.eu',
    'url': 'https://github.com/Steffo99/lokiunimore',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import re
import flask
import flask_sqlalchemy
import werkzeug.middleware.proxy_fix
import werkzeug.exceptions
import authlib.integrations.flask_client
import authlib.integrations.base_client
import logging
import requests

log = logging.getLogger(__name__)

from lokiunimore.config import config, FLASK_SECRET_KEY, SQLALCHEMY_DATABASE_URL
from lokiunimore.sql.tables import Base as TableDeclarativeBase
from lokiunimore.sql.tables import Account, MatrixUser
from lokiunimore.web.extensions.matrix_client import MatrixClientExtension


app = flask.Flask(__name__)
"""
The main :mod:`flask` application object.
"""

rp_app = werkzeug.middleware.proxy_fix.ProxyFix(app=app, x_for=1, x_proto=1, x_host=1, x_port=0, x_prefix=0)
"""
Reverse proxied instance of :data:`.app`, to use in production with a Caddy server.
"""

app.config.update({
    **config.proxies.resolve(),
    "SECRET_KEY": FLASK_SECRET_KEY.__wrapped__,
    "SQLALCHEMY_DATABASE_URI": SQLALCHEMY_DATABASE_URL.__wrapped__,
    "SQLALCHEMY_TRACK_MODIFICATIONS": False,
})


db = flask_sqlalchemy.SQLAlchemy(app=app, metadata=TableDeclarativeBase.metadata)
"""
:mod:`sqlalchemy` database engine, usable by the whole :data:`.app`.
"""

db.create_all()


oauth = authlib.integrations.flask_client.OAuth(app=app)
"""
OAuth2 :mod:`flask` extension installed on :data:`.app`.
"""

oauth.register(
    name="oidc",
    server_metadata_url=app.config["OIDC_CONFIGURATION_URL"],
    api_base_url=app.config["OIDC_API_BASE_URL"],
    client_kwargs={
        "scope": app.config["OIDC_SCOPES"]
    },
)


matrix = MatrixClientExtension(app=app)
"""
Synchronous Matrix client for use with Flask.
"""


### Setup the app routes

@app.route("/")
def page_root():
    return flask.render_template("root.html")


@app.route("/privacy")
def page_privacy():
    return flask.render_template("privacy.html")


@app.route("/matrix/<token>/")
def page_matrix_profile(token):
    user = db.session.query(MatrixUser).filter_by(token=token).first_or_404()
    return flask.render_template("matrix.html", user=user, token=token)


@app.route("/matrix/<token>/link")
def page_matrix_link(token):
    db.session.query(MatrixUser).filter_by(token=token).first_or_404()
    flask.session["matrix_token"] = token
    return oauth.oidc.authorize_redirect(flask.url_for("page_oidc_authorize", _external=True))


@app.route("/matrix/<token>/invite")
def page_matrix_invite(token):
    matrix_user: MatrixUser = db.session.query(MatrixUser).filter_by(token=token).first_or_404()

    log.debug(f"Sending private space invite to: {matrix_user.id}")
    matrix.room_invite(room_id=app.config["MATRIX_PRIVATE_SPACE_ID"], user_id=matrix_user.id)
    log.info(f"Sent private space invite to: {matrix_user.id}")

    return flask.redirect(flask.url_for("page_matrix_profile", token=token))


@app.route("/authorize")
def page_oidc_authorize():
    try:
        token = oauth.oidc.authorize_access_token()
    except werkzeug.exceptions.BadRequestKeyError:
        return flask.render_template(
            "error.html",
            when="""durante la verifica del login con Google""",
            details="""Mancano i parametri query necessari per effettuare l'autenticazione OAuth2.""",
            tip="""Rifai la procedura di connessione account da capo.<br>Se il problema persiste, inviami un'email a <a href="mailto:me@steffo.eu">me@steffo.eu</a>, e provvederò a risolvere il problema!"""
         )
    except authlib.integrations.base_client.errors.OAuthError:
        return flask.render_template(
            "error.html",
            when="""durante la verifica del login con Google""",
            details="""Qualcosa è inaspettatamente andato storto durante l'autenticazione OAuth2.""",
            tip="""Rifai la procedura di connessione account da capo.<br>Se il problema persiste, inviami un'email a <a href="mailto:me@steffo.eu">me@steffo.eu</a>, e provvederò a risolvere il problema!"""
        )

    # Not sure of why the nonce is now required or if I'm using it correctly
    account = oauth.oidc.parse_id_token(token=token, nonce=token["userinfo"]["nonce"])

    if not account.email_verified:
        return flask.render_template(
            "error.html",
            when="""durante la verifica del tuo account Google""",
            details="""L'email del tuo account Google non è verificata.""",
            tip="""Probabilmente hai effettuato l'accesso con l'account Google sbagliato.<br>Effettua il <a href="https://accounts.google.com/logout">logout da tutti i tuoi account Google</a> e riprova!"""
        ), 403

    # noinspection PyTypeChecker
    if not re.match(app.config["OIDC_EMAIL_REGEX"], account.email):
        return flask.render_template(
            "error.html",
            when="""durante la verifica del tuo account Google""",
            details="""Questo account Google non appartiene all'organizzazione <i>studenti@UniMoRe</i>.""",
            tip="""Probabilmente hai effettuato l'accesso con l'account Google sbagliato.<br>Effettua il <a href="https://accounts.google.com/logout">logout da tutti i tuoi account Google</a> e riprova!"""
        ), 403

    local_account = db.session.merge(Account(
        email=account.email,
        first_name=account.given_name,
        last_name=account.family_name,
    ))

    if matrix_token := flask.session.pop("matrix_token", None):
        matrix_user = db.session.query(MatrixUser).filter_by(token=matrix_token).first_or_404()
        matrix_user.account = local_account

        db.session.commit()

        return flask.redirect(flask.url_for("page_matrix_invite", token=matrix_token))

    else:
        return flask.render_template(
            "error.html",
            when="""durante il collegamento dei tuoi account""",
            details="""Non è stato possibile trovare l'account da collegare.""",
            tip="""Probabilmente hai rifiutato il cookie di sessione in cui viene salvato l'account locale da collegare. Assicurati che il tuo browser lo accetti, poi riprova!"""
        ), 403
